/*
   LDecomp version 1.0
   Copyright 2004 by ccexplore.  All rights reserved.

   The author hereby grants you non-exclusive, non-transferrable rights to:

     1) compile this code as is, then use and/or distribute the resulting
        program freely.

     2) distribute this code freely in its ORIGINAL form only, and without
        this code being part of another programming project

     3) use this code freely as part of your own programming project, PROVIDING
        that no parts of said programming project (including source files and
        binaries) will be distributed to the public

   Note: rights granted here are not transferrable to any other versions of
   this program/code.

   If you wish to incorporate any part of this code into your own programming
   project, and plan to distribute said project to the public, please contact
   the author at ccexplore@yahoo.com.  Until the author grants you explicit
   permission in writing, use of this code in your own programming projects is
   forbidden except for projects with no parts distributed to the public.

   ---------------------------------------

   Basically, if you're going to use this code in your own stuff, I want to
   make sure you give due credit to me.

   ---------------------------------------

   Disclaimer:  by using this source code and/or the program, you have agreed
   to indemnify me against any legal liability (except where prohibited by law)
   resulting from the use of this source code or the program.

   ---------------------------------------

   Ok, enough of this, I need a real lawyer to write this kind of stuff. ;p

 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>

typedef struct header_t {
  unsigned char bitcnt;
  unsigned char checksum;
  char unused1[2];
  unsigned char decompressed_size[2]; /* big-endian */
  char unused2[2];
  unsigned char compressed_size[2];  /* big-endian, size includes 10-byte header */
} header_t;

char ans[256];

void printheaderinfo(header_t *h, int c_size, int d_size, int count) {
  printf("header %d:\n", count);
  printf("\tbitcnt = %d\n", h->bitcnt);
  printf("\tchecksum = 0x%02x\n", h->checksum);
  printf("\tsize of decompressed data = %d bytes\n", d_size);
  printf("\tsize of compressed data (includes header) = %d bytes\n", c_size);
}

typedef struct decompressor_state_t {
  unsigned char bitcnt;
  unsigned char curbits;
  const unsigned char *cdata;
  int cptr;
  unsigned char *ddata;
  int dptr;
  int checksum;
} decompressor_state_t;

int getnextbits(int n, decompressor_state_t *state) {
  int result = 0;
  do {
    if (--state->bitcnt == 0) {
      state->cptr--;
      state->curbits = state->cdata[state->cptr];
      state->checksum ^= state->curbits;
      state->bitcnt = 8;
    }
    result <<= 1;
    result |= (state->curbits & 1);
    state->curbits >>= 1;
  } while(--n > 0);
  return result;
}

void copy_prev_data(int blocklen, int offset_size, decompressor_state_t *state) {
  int offset = getnextbits(offset_size, state);
  int i;

  for(i = 0; i < blocklen; i++) {
    state->dptr--;
    state->ddata[state->dptr] = state->ddata[state->dptr + offset + 1];
  }
}

void dump_data(int numbytes, decompressor_state_t *state) {
  do {
    state->dptr--;
    state->ddata[state->dptr] = (unsigned char)getnextbits(8, state);
  } while (--numbytes > 0);
}

int decompress(const unsigned char *cdata, unsigned char *ddata,
               unsigned char bitcnt, int csize, int dsize) {
  decompressor_state_t state;

  state.bitcnt = bitcnt + 1;
  state.cptr = csize - 1;
  state.dptr = dsize;
  state.cdata = cdata;
  state.ddata = ddata;
  state.curbits = cdata[state.cptr];
  state.checksum = state.curbits;

  do {
    if (getnextbits(1, &state) == 1) {
      switch (getnextbits(2, &state)) {
        case 0:
          copy_prev_data(3, 9, &state);
          break;
        case 1:
          copy_prev_data(4, 10, &state);
          break;
        case 2:
          copy_prev_data(getnextbits(8, &state) + 1, 12, &state);
          break;
        case 3:
          dump_data(getnextbits(8, &state) + 9, &state);
          break;
      }
    }
    else {
      switch (getnextbits(1, &state)) {
        case 0:
          dump_data(getnextbits(3, &state) + 1, &state);
          break;
        case 1:
          copy_prev_data(2, 8, &state);
          break;
      }
    }
  } while (state.dptr > 0);

  return state.checksum;
}

void save(const unsigned char *data, int size) {
  FILE *savefile = NULL;

  do {
    printf("Enter filename: ");
    scanf("%*[\n]%255[^\n]", ans);
    savefile = fopen(ans, "wb");
    if (savefile == NULL) {
      printf("file %s cannot be opened for writing!  Try again (y/n)? ", ans);
      scanf("%1s", ans);
      if (ans[0] == 'n' || ans[0] == 'N')
        return;
      else
        continue;
    }
    else {
      printf("%u bytes written.\n", (unsigned)fwrite(data, 1, size, savefile));
      fclose(savefile);
      return;
    }
  } while(1);
}

int main(int argc, char *argv[]) {
  int count = 1;
  FILE *datfile = NULL;
  header_t header;
  int compressed_size;
  int decompressed_size;
  unsigned char *compress_data = NULL;
  unsigned char *decompress_data = NULL;
  int computed_checksum;
  long totalnumbytes;

  if (sizeof(header_t) != 10) {  /* just in case */
    printf("ERROR: please recompile program to use BYTE alignment for struct members\n");
    exit(EXIT_FAILURE);
  }

  printf("--------------------------------\n");
  printf("This is ldecomp version 1.0\n");
  printf("Copyright 2004 by ccexplore.  This program is freeware.\n");
  printf("By using this program, you have agreed to the terms and conditions\n");
  printf("set forth in the copyright and disclaimer sections of the included\n");
  printf("source file.\n");
  printf("--------------------------------\n\n");

  if (argc < 2) {
    printf("Enter filename of .dat file to examine (please include the \".dat\"): ");
    scanf("%255[^\n]", ans);
  }
  else {
    strcpy(ans, argv[1]);
  }

  datfile = fopen(ans, "rb");
  if (datfile == NULL) {
    printf("file %s cannot be opened, aborting...\n", ans);
    exit(EXIT_FAILURE);
  }

  fseek(datfile, 0, SEEK_END);
  totalnumbytes = ftell(datfile);
  fseek(datfile, 0, SEEK_SET);

  do {
    fread(&header, 10, 1, datfile);
    totalnumbytes -= 10;
    compressed_size = header.compressed_size[1] + header.compressed_size[0] * 256;
    decompressed_size = header.decompressed_size[1] + header.decompressed_size[0] * 256;
    printheaderinfo(&header, compressed_size, decompressed_size, count);
    count++;

    compressed_size -= 10;

    compress_data = malloc(compressed_size);
    decompress_data = malloc(decompressed_size);
    if (compress_data == NULL || decompress_data == NULL) {
      printf("not enough memory to extract data!\n");
      goto nextheader;
    }

    fread(compress_data, 1, compressed_size, datfile);
    totalnumbytes -= compressed_size;
    computed_checksum = decompress(compress_data, decompress_data, header.bitcnt,
                                   compressed_size, decompressed_size);
    if (header.checksum != (unsigned char)computed_checksum) {
      printf("warning: computed checksum of 0x%02x does not "
             "match the checksum stored in header\n", computed_checksum);
    }

    printf("Save decompressed data (y/n)? ");
    scanf("%1s", ans);
    if (ans[0] == 'Y' || ans[0] == 'y')
      save(decompress_data, decompressed_size);

nextheader:
    free(compress_data);
    free(decompress_data);
    printf("\nContinue to next header (y/n)? ");
    scanf("%1s", ans);
  } while(ans[0] != 'n' && ans[0] != 'N' && totalnumbytes > 0);

  if (totalnumbytes == 0)
    printf("End of file reached.  ");
  else if (totalnumbytes < 0)
    printf("warning: we\'ve read past end of file, last decompressed section might be corrupted\n");

  printf("Goodbye!\n");
  fclose(datfile);
  return 0;
}
